/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * (c) Copyright 1998 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmNddsResponder.h
 *
 * ABSTRACT: Functions for responding to TCM invocations from ControlShell,
 *           via NDDS
 *
 * EXPORTS:
 *
 * $Revision: 1.3 $
 * $Date: 1998/04/21 17:58:23 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 * $Log: tcmNddsResponder.h,v $
 * Revision 1.3  1998/04/21 17:58:23  reids
 * Changes to make stimulus/response messages be reliable productions in NDDS
 *
 * Revision 1.2  98/03/18  14:41:52  reids
 * Changes to support both SunOS and Solaris versions.
 * 
 * Revision 1.1  98/01/02  00:21:12  reids
 * Initial attempt to integrate TCM and ControlShell via NDDS.
 *   A bit kludgy, since I don't really understand how to handle a series
 *   of productions -- really want them to be queued, reliably...
 * 
 **************************************************************************/

#ifndef INCtcmNddsResponder
#define INCtcmNddsResponder

#include "tcmResponder.h"
#include "NDDS.h"

const unsigned int TCM_NDDS_DEFAULT_POLL_PERIOD = 250;/* msecs */

const unsigned int NDDS_RELIABLE_QUEUE_SIZE = 10;

/*** Defined in tcmNddsResponse.h ***
typedef struct TCM_NDDS_Response_Type {
  int id;
  String response;
} TCM_NDDS_Response_Type;
****************************************************************/
#include "tcmNddsResponse.h"
typedef TCM_NDDS_Response_Type *TCM_NDDS_Response_Ptr;

/*** Defined in tcmNddsStimulus.h ***
typedef struct TCM_NDDS_Stimulus_Type {
  int id;
  String stimulus;
  int numResponses;
  TCM_CS_Response_Type responses[MAX_RESPONSES];
} TCM_NDDS_Stimulus_Type;
****************************************************************/
#include "tcmNddsStimulus.h"
typedef TCM_NDDS_Stimulus_Type *TCM_NDDS_Stimulus_Ptr;

typedef class tcmNddsResponderClass *tcmNddsResponder;

class tcmNddsResponderClass : public tcmResponderClass
{
 public:
  tcmNddsResponderClass (String stimulus, int refId, int maxResponseTableSize)
    : tcmResponderClass(stimulus, NULL, maxResponseTableSize)
    { _refId = refId; }
    
  virtual boolean respond (String context, int returnCode) const;

  virtual boolean missingStimulus (String stimulus) const;

 protected:
  int _refId;
};

/****************************************************************
 *
 * Public Functions
 *
 ****************************************************************/

/*****************************************************************
 * Need to invoke this within CS at startup, to initialize all the 
 * things needed for CS to respond to TCM commands via NDDS.
 *******************************************************************/
TCM_Return_Type tcmNddsResponderInit (CSFsm fsm);

/*******************************************************************
 * This is invoked at the CS side, to handle stimulus messages from TCM.
 * It should be called periodically.
 *******************************************************************/
void tcmNddsResponderPoll (void);

/*******************************************************************
 * This is invoked at the CS side, to send a response back to TCM.
 * The "refId" is associated with a task tree node, and the response
 *  is either SUCCESS_RESPONSE, or an arbitrary string denoting failure.
 *******************************************************************/
void tcmNddsRespond (int refId, String response);

/*******************************************************************
 * Need to invoke this within the TCM process at startup, to initialize
 *  all the things needed for TCM to command CS via NDDS.
 *******************************************************************/
TCM_Return_Type tcmNddsSenderInit (void);

/*******************************************************************
 * This is invoked at the TCM side, to handle response messages from CS.
 * It should be called periodically.
 *******************************************************************/
void tcmNddsSenderPoll (void);

/*******************************************************************
 *
 * Send a stimulus to CS via NDDS and tell CS to set up the intention
 *   to await response(s) to that stimulus (which will then be sent
 *   back via NDDS).
 ****************************************************************/
TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode,
				 String stimulus, int numResponses,
				 TCM_CS_Response_Type *responses);

/*******************************************************************
 *
 * Send a stimulus to CS via NDDS and tell CS to set up the intention to
 *   await response(s) to that stimulus (which will then be sent back via NDDS).
 *
 * The variable arguments are triples of the form:
 *  <context, retcode, response-string>, where "context" is the name of the
 *  state (a string) in which the response is enabled, "retcode" is the
 *  transition return code that triggers this response, and "response-string"
 *  is what to send back to TCM (either SUCCESS_RESPONSE or an arbitrary
 *  string, signifying failure).  The end of the list is indicated by the
 *  special symbol "END_RESPONSES" (defined in tcmResponder.h).
 *
 * eg, tcmNddsSendStimulus(node, "StartBB", "StartTraj", 1, "NoTrajectory",
 *                         "Completed", 0, SUCCESS_RESPONSE, END_RESPONSES);
 *******************************************************************/
TCM_Return_Type tcmNddsSendStimulus (TCM_Task_Tree_Ref parentNode,
				     String stimulus, ...);

#endif // INCtcmNddsResponder
