/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * (c) Copyright 1997 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmResponder.h
 *
 * ABSTRACT: Functions for responding to TCM invocations from ControlShell
 *
 * EXPORTS:
 *
 * $Revision: 1.3 $
 * $Date: 1998/03/18 14:41:54 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 * $Log: tcmResponder.h,v $
 * Revision 1.3  1998/03/18 14:41:54  reids
 * Changes to support both SunOS and Solaris versions.
 *
 * Revision 1.2  98/01/02  00:21:16  reids
 * Initial attempt to integrate TCM and ControlShell via NDDS.
 *   A bit kludgy, since I don't really understand how to handle a series
 *   of productions -- really want them to be queued, reliably...
 * 
 * Revision 1.1  97/11/21  14:05:09  reids
 * First release of TCM -- seems to be a stable version
 **************************************************************************/

#ifndef INCtcmResponder
#define INCtcmResponder

#include "cs.h"
#include "CSFsm.h"
#include "CSFsmResponder.h"
#include "tcm.h"

const int MAX_RESPONSES = 10;

const String SUCCESS_RESPONSE = "Success";

typedef struct TCM_CS_Response_Type {
  String context;
  int returnCode;
  String responseString;

  TCM_CS_Response_Type() { context = responseString = ""; returnCode = 0; }
  ~TCM_CS_Response_Type() { freeStrings(); }
  void set (String theContext, int retCode, String response) {
    freeStrings();
    context = strdup(theContext); returnCode = retCode;
    responseString = strdup(response);
  }

  private:
  void freeStrings(void) {
    if (context && strlen(context) > 0) delete[] context;
    if (responseString && strlen(responseString) > 0) delete[] responseString;
  }
} TCM_CS_Response_Type, *TCM_CS_Response_Ptr;

class ResponseTableClass
{
 public:
  ResponseTableClass (int maxElements);
  ResponseTableClass (ResponseTableClass &table);
  ~ResponseTableClass() { if (_table) delete[] _table; }

  void add (String context, int returnCode, String responseString);
  String getResponse (String context, int returnCode) const;

 private:
  int _maxElements;
  int _numElements;
  TCM_CS_Response_Type *_table; // Variable length array
};

typedef class tcmResponderClass *tcmResponder;

class tcmResponderClass : public CSFsmResponderClass
{
 public:
  tcmResponderClass(String stimulus, TCM_Task_Tree_Ref treeRef,
		    int maxResponseTableSize) :
    CSFsmResponderClass(stimulus),
    _treeRef(treeRef), _responseTable(maxResponseTableSize) {}

  virtual boolean respond (String context, int returnCode) const;
  virtual boolean missingStimulus (String stimulus) const;

  void addResponse (String context, int returnCode, String responseString)
    { _responseTable.add(context, returnCode, responseString); }

 protected:
  TCM_Task_Tree_Ref _treeRef;
  ResponseTableClass _responseTable;
};

/***
 *
 * Public function for sending a stimulus to CS and setting up the
 *  intention to await response(s) to that stimulus (all within a
 *  single process).
 *
 ***/
TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode, CSFsm fsm,
				 String stimulus, int numResponses,
				 TCM_CS_Response_Type *responses);

/***
 *
 * Public function for sending a stimulus to CS and setting up the
 *  intention to await response(s) to that stimulus (all within a
 *  single process).
 *
 * The variable arguments are triples of the form:
 *  <context, retcode, response-string>, where "context" is the name of the
 *  state (a string) in which the response is enabled, "retcode" is the
 *  transition return code that triggers this response, and "response-string"
 *  is what to send back to TCM (either SUCCESS_RESPONSE or an arbitrary
 *  string, signifying failure).  The end of the list is indicated by the
 *  special symbol "END_RESPONSES".
 *
 *  e.g. tcmSendStimulus(node, fsm, "StartBB", "StartTraj", 1, "NoTrajectory",
 *                       "Completed", 0, SUCCESS_RESPONSE, END_RESPONSES);
 ***/

#define END_RESPONSES ((void *)-1)

TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode, CSFsm fsm,
				 String stimulus, ...);
				 
#endif // INCtcmResponder
