
#ifndef TDL_DEF_H
#define TDL_DEF_H


  /*
   * Class TDL exists primarily for use as a namespace,
   *  and should NEVER be constructed.
   */
class TDL
{
public:
  enum  REPORTING_LEVELS
  {
    NO_REPORTING                = 0x0000,

	/* Basic stuff that you almost always want... */
    SPAWN_FAILURE_OUTSIDE_TASK  = 0x0001,
    EVERYTHING_BASIC            = SPAWN_FAILURE_OUTSIDE_TASK,

	/* Informational Reporting Verbosity... */
    VERBOSE_BRIEF               = 0x0002,
    VERBOSE_FULL                = 0x0004,
    VERBOSE                     = VERBOSE_BRIEF | VERBOSE_FULL,

	/* Everything up above all OR'ed together. (With room for additions) */
    EVERYTHING                  = 0x00ff,

	/* Debugging reports are not included by default. */
    TDL_DEBUG                   = 0x0100,
    TDL_DEBUG_TIME              = 0x0200,

	/* Everything -- including debugging reports! */
    ALL                         = 0xffff
  };

protected:
  static u_int4    reportingLevel;  /* = _TDL_REPORTING_LEVEL *
				     *    or TDL::EVERYTHING  */
  static BOOLEAN   closeLogStream;  /* = FALSE */
  static ostream * logStream;       /* = & cerr */

public:
  static ostream &  getLogStream()  { return * TDL::logStream; }

  static void setLogStream ( ostream & theOstream )
  {
    if ( TDL::closeLogStream == TRUE )
      delete TDL::logStream;

    TDL::closeLogStream = FALSE;
    TDL::logStream      = & theOstream;
  }

  static void setLogStream ( const char * theFilename )
  {
    if ( TDL::closeLogStream == TRUE )
      delete TDL::logStream;

    TDL::closeLogStream = TRUE;
    TDL::logStream      = new ofstream ( theFilename );
  }


  static u_int4 getReportingLevel ( )   { return TDL::reportingLevel; }
  static void   setReportingLevel ( u_int4 theReportingLevel )
		  { TDL::reportingLevel = theReportingLevel; }

  static void   addReportingLevel    ( u_int4 theReportingLevel )
		  { setReportingLevel (       getReportingLevel() 
					  |   theReportingLevel   ); }

  static void   removeReportingLevel ( u_int4 theReportingLevel )
		  { setReportingLevel (       getReportingLevel() 
					| ( ~ theReportingLevel ) ); }


  static inline int getIsReporting ( u_int4 theReportingLevel )
	 { return ( ( theReportingLevel & TDL::getReportingLevel() ) != 0 ); }


  static inline BOOLEAN shouldReportError ( TCM_Return_Type theResult,
					    int             theErrorLevel )
  {
    return (   ( theResult != TCM_Ok                   )
	    && ( TDL::getIsReporting ( theErrorLevel ) ) );
  }

	/* Note: Return TRUE if we Succeed, FALSE for Failure */
  static inline BOOLEAN testAndHandleFailure (
					TCM_Return_Type theResult,
					int             theErrorLevel,
					const char *    theErrorMessage1,
					const char *    theErrorMessage2 = "",
					const char *    theErrorMessage3 = "" )
  {
    if ( TDL::shouldReportError ( theResult, theErrorLevel ) )
    {
      getLogStream()
	<< theErrorMessage1 << theErrorMessage2 << theErrorMessage3 << endl;
    }
    return ( theResult == TCM_Ok );
  }

	/* Note: Return TRUE if we Succeed, FALSE for Failure */
  static inline BOOLEAN verifyAllocation (
				 const char *               theLocation,
				 const TCM_Task_Tree_Ref &  theTCMTaskTreeRef )
  {
    return testAndHandleFailure (
	    (  ( theTCMTaskTreeRef . operator*()  ==  (Task_Tree_Node *) NULL )
	      ? TCM_Error : TCM_Ok ),
	    TDL::SPAWN_FAILURE_OUTSIDE_TASK,
	    theLocation,
	    "  Failure:  Allocation/Access Failure.  "
	    "(Perhaps this task has been destroyed?)"  );

  }

	/* Note: Return TRUE if we Succeed, FALSE for Failure */
  static inline BOOLEAN verifySetAction  ( const char *     theLocation,
					   TCM_Return_Type  theResult   )
  {
    return testAndHandleFailure ( theResult,
				  TDL::SPAWN_FAILURE_OUTSIDE_TASK,
				  theLocation,
				  "  Failure:  _TDL_SetAction_..." );
  }

	/* Note: Return TRUE if we Succeed, FALSE for Failure */
  static inline BOOLEAN verifyInsertNode ( const char *     theLocation,
					   TCM_Return_Type  theResult   )
  {
    return testAndHandleFailure ( theResult,
				  TDL::SPAWN_FAILURE_OUTSIDE_TASK,
				  theLocation,
				  "  Failure:  TCM_InsertNode" );
  }

	/* Note: Return TRUE if we Succeed, FALSE for Failure */
  static inline BOOLEAN verifyConstraint ( const char *     theLocation,
					   const char *     theConstraint,
					   TCM_Return_Type  theResult     )
  {
    return testAndHandleFailure ( theResult,
				  TDL::SPAWN_FAILURE_OUTSIDE_TASK,
				  theLocation,
				  "  Constraint Failure:  ",
				  theConstraint );
  }

  static inline TCM_Task_Tree_Ref doSetOnTerminateTaskAction (
				  const char *              theLocation,
				  const TCM_Task_Tree_Ref & theTaskReference,
				  _Action *                 theTaskAction    )
  {
    if (   TDL::verifyAllocation ( theLocation, theTaskReference )
	&& TDL::verifySetAction  ( theLocation,
				   TCM_SetActualAction ( theTaskReference,
							 theTaskAction    ) )
	)
      return theTaskReference;
    else
      return NULL;
  }


	/* Returns TRUE for direct instances [SPAWN foo();],   *
	 * both Monitor and Monitor-child instances ["ACT-"],  *
	 * and SPAWN-with-WAIT macro instances ["-auto,wait"]. */
  static BOOLEAN checkForTaskInstanceOf ( const char * theActualTaskName,
					  const char * theNameToCheck );


private:
  TDL(){} /* Disable construction of this class */

}; /* class TDL */


#endif /* TDL_DEF_H */

